#ifndef lint
static char sccsid[] = "@(#)tiff_dir.c	1.22 2/19/89";
#endif

/*
 * Copyright (c) 1988 by Sam Leffler.
 * All rights reserved.
 *
 * This file is provided for unrestricted use provided that this
 * legend is included on all tape media and as a part of the
 * software program in whole or part.  Users may copy, modify or
 * distribute this file at will.
 */

/*
 * TIFF Library.
 *
 * Directory Support Routines.
 */
#include "tiffio.h"

#define	ReadOK(fd, buf, size)	(read(fd, (char *)buf, size) == size)
#define	WriteOK(fd, buf, size)	(write(fd, (char *)buf, size) == size)
#define	SeekOK(fd, off, whence)	(lseek(fd, (long)off, whence) == (long)off)

#define	howmany(x, y)	(((x)+((y)-1))/(y))
#define	roundup(x, y)	((((x)+((y)-1))/(y))*(y))

#define	FieldSet(fields, field)		(fields[field/32] & (1<<(field&0x1f)))
#define	ResetFieldBit(fields, field)	(fields[field/32] &= ~(1<<(field&0x1f)))

typedef	struct {
	u_short	field_tag;		/* field's tag */
	short	field_count;		/* expected item count */
	TIFFDataType field_type;	/* type of associated data */
	u_short	field_bit;		/* bit in fieldsset bit vector */
	char	*field_name;		/* ASCII name */
} TIFFFieldInfo;

/*
 * NB: This array is assumed to be sorted by tag.
 */
static	TIFFFieldInfo FieldInfo[] = {
    { TIFFTAG_SUBFILETYPE,	 1, SHORT,	FIELD_SUBFILETYPE,
      "SubfileType" },
    { TIFFTAG_OSUBFILETYPE,	 1, SHORT,	FIELD_SUBFILETYPE,
      "OldSubfileType" },
    { TIFFTAG_IMAGEWIDTH,	 1, SHORT,	FIELD_IMAGEDIMENSIONS,
      "ImageWidth" },
    { TIFFTAG_IMAGELENGTH,	 1, SHORT,	FIELD_IMAGEDIMENSIONS,
      "ImageLength" },
    { TIFFTAG_BITSPERSAMPLE,	-1, SHORT,	FIELD_BITSPERSAMPLE,
      "BitsPerSample" },
    { TIFFTAG_COMPRESSION,	 1, SHORT,	FIELD_COMPRESSION,
      "Compression" },
    { TIFFTAG_PHOTOMETRIC,	 1, SHORT,	FIELD_PHOTOMETRIC,
      "PhotometricInterpretation" },
    { TIFFTAG_THRESHHOLDING,	 1, SHORT,	FIELD_THRESHHOLDING,
      "Threshholding" },
    { TIFFTAG_CELLWIDTH,	 1, SHORT,	(u_short)-1,
      "CellWidth" },
    { TIFFTAG_CELLLENGTH,	 1, SHORT,	(u_short)-1,
      "CellLength" },
    { TIFFTAG_FILLORDER,	 1, SHORT,	(u_short)-1,
      "FillOrder" },
    { TIFFTAG_DOCUMENTNAME,	-1, ASCII,	FIELD_DOCUMENTNAME,
      "DocumentName" },
    { TIFFTAG_IMAGEDESCRIPTION,	-1, ASCII,	FIELD_IMAGEDESCRIPTION,
      "ImageDescription" },
    { TIFFTAG_MAKE,		-1, ASCII,	FIELD_MAKE,
      "Make" },
    { TIFFTAG_MODEL,		-1, ASCII,	FIELD_MODEL,
      "Model" },
    { TIFFTAG_STRIPOFFSETS,	-1, LONG,	FIELD_STRIPOFFSETS,
      "StripOffsets" },
    { TIFFTAG_STRIPOFFSETS,	-1, SHORT,	FIELD_STRIPOFFSETS,
      "StripOffsets" },
    { TIFFTAG_ORIENTATION,	 1, SHORT,	FIELD_ORIENTATION,
      "Orientation" },
    { TIFFTAG_SAMPLESPERPIXEL,	 1, SHORT,	FIELD_SAMPLESPERPIXEL,
      "SamplesPerPixel" },
    { TIFFTAG_ROWSPERSTRIP,	-1, LONG,	FIELD_ROWSPERSTRIP,
      "RowsPerStrip" },
    { TIFFTAG_ROWSPERSTRIP,	-1, SHORT,	FIELD_ROWSPERSTRIP,
      "RowsPerStrip" },
    { TIFFTAG_STRIPBYTECOUNTS,	-1, LONG,	FIELD_STRIPBYTECOUNTS,
      "StripByteCounts" },
    { TIFFTAG_MINSAMPLEVALUE,	-1, SHORT,	FIELD_MINSAMPLEVALUE,
      "MinSampleValue" },
    { TIFFTAG_MAXSAMPLEVALUE,	-1, SHORT,	FIELD_MAXSAMPLEVALUE,
      "MaxSampleValue" },
    { TIFFTAG_XRESOLUTION,	 1, RATIONAL,	FIELD_RESOLUTION,
      "XResolution" },
    { TIFFTAG_YRESOLUTION,	 1, RATIONAL,	FIELD_RESOLUTION,
      "YResolution" },
    { TIFFTAG_PLANARCONFIG,	 1, SHORT,	FIELD_PLANARCONFIG,
      "PlanarConfiguration" },
    { TIFFTAG_PAGENAME,		-1, ASCII,	FIELD_PAGENAME,
      "PageName" },
    { TIFFTAG_XPOSITION,	 1, RATIONAL,	FIELD_POSITION,
      "XPosition" },
    { TIFFTAG_YPOSITION,	 1, RATIONAL,	FIELD_POSITION,
      "YPosition" },
    { TIFFTAG_FREEOFFSETS,	-1, LONG,	(u_short)-1,
      "FreeOffsets" },
    { TIFFTAG_FREEBYTECOUNTS,	-1, LONG,	(u_short)-1,
      "FreeByteCounts" },
    { TIFFTAG_GRAYRESPONSEUNIT,	 1, SHORT,	FIELD_GRAYRESPONSEUNIT,
      "GrayResponseUnit" },
    { TIFFTAG_GRAYRESPONSECURVE,-1, SHORT,	FIELD_GRAYRESPONSECURVE,
      "GrayResponseCurve" },
    { TIFFTAG_GROUP3OPTIONS,	 1, LONG,	FIELD_GROUP3OPTIONS,
      "Group3Options" },
    { TIFFTAG_GROUP4OPTIONS,	 1, LONG,	FIELD_GROUP4OPTIONS,
      "Group4Options" },
    { TIFFTAG_RESOLUTIONUNIT,	 1, SHORT,	FIELD_RESOLUTIONUNIT,
      "ResolutionUnit" },
    { TIFFTAG_PAGENUMBER,	 2, SHORT,	FIELD_PAGENUMBER,
      "PageNumber" },
    { TIFFTAG_COLORRESPONSEUNIT, 1, SHORT,	FIELD_COLORRESPONSEUNIT,
      "ColorResponseUnit" },
    { TIFFTAG_COLORRESPONSECURVE,-1,SHORT,	FIELD_COLORRESPONSECURVE,
      "ColorResponseCurve" },
    { TIFFTAG_SOFTWARE,		-1, ASCII,	FIELD_SOFTWARE,
      "Software" },
    { TIFFTAG_DATETIME,		-1, ASCII,	FIELD_DATETIME,
      "DateTime" },
    { TIFFTAG_ARTIST,		-1, ASCII,	FIELD_ARTIST,
      "Artist" },
    { TIFFTAG_HOSTCOMPUTER,	-1, ASCII,	FIELD_HOSTCOMPUTER,
      "HostComputer" },
    { TIFFTAG_PREDICTOR,	 1, SHORT,	FIELD_PREDICTOR,
      "Predictor" },
    { TIFFTAG_COLORMAP,		-1, SHORT,	FIELD_COLORMAP,
      "ColorMap" },
/* begin nonstandard tags */
    { TIFFTAG_MATTEING,		 1, SHORT,	FIELD_MATTEING,
      "Matteing" },
};
#define	NFIELDINFO	(sizeof (FieldInfo) / sizeof (FieldInfo[0]))

#define	IGNORE	0		/* tag placeholder used below */

static int datawidth[] = {
    1,	/* nothing */
    1,	/* BYTE */
    1,	/* ASCII */
    2,	/* SHORT */
    4,	/* LONG */
    8,	/* RATIONAL */
    4,	/* IEEEFLOAT */
    8,	/* IEEEDOUBLE */
};

static	double TIFFFetchRational();
#ifdef notdef
static	double TIFFFetchIEEE();
#endif
static	TIFFFieldInfo *FieldWithTag();
static	TIFFDirectory *TIFFNewDirectory();
extern	char *malloc();

/*
 * Read the next TIFF directory from a file
 * and convert it to the internal format.
 * We read directories sequentially.
 */
TIFFReadDirectory(tif)
	TIFF *tif;
{
	register TIFFDirEntry *dp;
	register int n;
	register TIFFDirectory *td = NULL;
	TIFFDirEntry *dir = NULL, *bits;
	long v;
	TIFFFieldInfo *fip;
	short dircount;
	char *cp;

	tif->tif_diroff = tif->tif_nextdiroff;
	if (tif->tif_diroff == 0)		/* no more directories */
		return (0);
	if (!SeekOK(tif->tif_fd, tif->tif_diroff, L_SET)) {
		TIFFError(tif->tif_name, "Seek error accessing TIFF directory");
		return (0);
	}
	if (!ReadOK(tif->tif_fd, &dircount, sizeof (short))) {
		TIFFError(tif->tif_name, "Can not read TIFF directory count");
		return (0);
	}
	if (tif->tif_flags & TIFF_SWAB)
		TIFFSwabShort(&dircount);
	dir = (TIFFDirEntry *)malloc(dircount * sizeof (TIFFDirEntry));
	if (dir == NULL) {
		TIFFError(tif->tif_name, "No space to read TIFF directory");
		return (0);
	}
	if (!ReadOK(tif->tif_fd, dir, dircount * sizeof (TIFFDirEntry))) {
		TIFFError(tif->tif_name, "Can not read TIFF directory");
		goto bad;
	}
	/*
	 * Read offset to next directory for sequential scans.
	 */
	if (!ReadOK(tif->tif_fd, &tif->tif_nextdiroff, sizeof (long)))
		tif->tif_nextdiroff = 0;
	if (tif->tif_flags & TIFF_SWAB)
		TIFFSwabLong(&tif->tif_nextdiroff);

	tif->tif_flags &= ~TIFF_BEENWRITING;	/* reset before new dir */
	/*
	 * Setup default value and then make a pass over
	 * the fields to check type and tag information,
	 * and to extract info required to size data
	 * structures.  A second pass is made afterwards
	 * read in everthing not taken in the first pass.
	 */
	td = TIFFNewDirectory(tif);
	if (td == NULL)
		goto bad;
	td->td_samplesperpixel = 1;
	td->td_rowsperstrip = 0xffffffff;
	td->td_bitspersample = 1;
	td->td_threshholding = THRESHHOLD_BILEVEL;
	td->td_orientation = ORIENTATION_TOPLEFT;
	td->td_resolutionunit = RESUNIT_INCH;
	td->td_grayresponseunit = GRAYRESPONSEUNIT_100S;
	td->td_colorresponseunit = COLORRESPONSEUNIT_100S;
#ifdef EAHACK
	/*
	 * Electronic Arts writes gray-scale TIFF files
	 * without a PlanarConfiguration directory entry.
	 * Thus we setup a default value here, even though
	 * the TIFF spec says there is no default value.
	 */
	TIFFSetField(tif, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
#endif EAHACK
	for (fip = FieldInfo, dp = dir, n = dircount; n > 0; n--, dp++) {
		if (tif->tif_flags & TIFF_SWAB) {
			TIFFSwabArrayOfShort(&dp->tdir_tag, 2);
			TIFFSwabArrayOfLong(&dp->tdir_count, 2);
		}
		/*
		 * Find the field information entry for this tag.
		 */
		while (fip < &FieldInfo[NFIELDINFO] &&
		    fip->field_tag < dp->tdir_tag)
			fip++;
		if (fip >= &FieldInfo[NFIELDINFO] ||
		    fip->field_tag != dp->tdir_tag) {
			/*
			 * Unknown field, should this be fatal?
			 */
			TIFFWarning(tif->tif_name,
			    "ignoring unknown field with tag %d (0x%x)",
				dp->tdir_tag, dp->tdir_tag);
			continue;
		}
		/*
		 * Null out old tags that we ignore.
		 */
		if (fip->field_bit == (u_short)-1) {
			dp->tdir_tag = IGNORE;
			continue;
		}
		/*
		 * Check data type.
		 */
		while ((TIFFDataType)dp->tdir_type != fip->field_type) {
			fip++;
			if (fip >= &FieldInfo[NFIELDINFO] ||
			    fip->field_tag != dp->tdir_tag) {
				TIFFError(tif->tif_name,
				    "Wrong data type %d for field \"%s\"",
				    dp->tdir_type, fip->field_name);
				goto bad;
			}
		}
		switch (dp->tdir_tag) {
		case TIFFTAG_STRIPOFFSETS:
		case TIFFTAG_STRIPBYTECOUNTS:
			TIFFSetFieldBit(td, fip->field_bit);
			break;
		case TIFFTAG_IMAGELENGTH:
		case TIFFTAG_PLANARCONFIG:
		case TIFFTAG_ROWSPERSTRIP:
		case TIFFTAG_SAMPLESPERPIXEL:
			if (!TIFFSetField(tif, dp->tdir_tag,
			  TIFFExtractData(tif, dp->tdir_type, dp->tdir_offset)))
				goto bad;
			break;
		}
	}

	/*
	 * Allocate directory structure and setup defaults.
	 */
	if (!TIFFFieldSet(td, FIELD_IMAGEDIMENSIONS)) {
		TIFFError(tif->tif_name,
		    "TIFF directory is missing required \"ImageLength\" field");
		goto bad;
	}
	if (!TIFFFieldSet(td, FIELD_PLANARCONFIG)) {
		TIFFError(tif->tif_name,
	    "TIFF directory is missing required \"PlanarConfiguration\" field");
		goto bad;
	}
	if (!TIFFFieldSet(td, FIELD_STRIPOFFSETS)) {
		TIFFError(tif->tif_name,
	    "TIFF directory is missing required \"StripOffsets\" field");
		goto bad;
	}
	if (!TIFFFieldSet(td, FIELD_STRIPBYTECOUNTS)) {
#ifndef BYTECOUNTHACK
		TIFFError(tif->tif_name,
	    "TIFF directory is missing required \"StripByteCounts\" field");
		goto bad;
#endif
	}
	td->td_nstrips = (td->td_rowsperstrip == 0xffffffff ? 1:
	    howmany(td->td_imagelength, td->td_rowsperstrip));
	if (td->td_planarconfig == PLANARCONFIG_SEPARATE)
		td->td_nstrips *= td->td_samplesperpixel;

	/*
	 * Second pass: extract other information.
	 *
	 * Should do a better job of verifying values.
	 */
	for (dp = dir, n = dircount; n > 0; n--, dp++)
		switch (dp->tdir_tag) {
		case IGNORE:			/* marked above */
			break;
		case TIFFTAG_YPOSITION:
		case TIFFTAG_XPOSITION:
		case TIFFTAG_YRESOLUTION:
		case TIFFTAG_XRESOLUTION:
			if (!TIFFSetField(tif, dp->tdir_tag,
			    TIFFFetchRational(tif, dp)))
				goto bad;
			break;
		case TIFFTAG_DOCUMENTNAME:
		case TIFFTAG_IMAGEDESCRIPTION:
		case TIFFTAG_MAKE:
		case TIFFTAG_MODEL:
		case TIFFTAG_PAGENAME:
			cp = malloc(dp->tdir_count);
			if (cp == NULL)
				goto bad;
			if (!TIFFSetField(tif, dp->tdir_tag, cp) ||
			    !TIFFFetchData(tif, dp, cp))
				goto bad;
			break;
		case TIFFTAG_COMPRESSION:
			/*
			 * The 5.0 spec says the compression tag has
			 * one value, while earlier specs say it has
			 * one value per sample.  Because of this, we
			 * accept the tag if one value is supplied.
			 */
			if (dp->tdir_count == 1) {
				v = TIFFExtractData(tif,
				    dp->tdir_type, dp->tdir_offset);
				if (!TIFFSetField(tif, TIFFTAG_COMPRESSION, v))
					goto bad;
				break;
			}
			/* fall thru... */
		case TIFFTAG_MINSAMPLEVALUE:
		case TIFFTAG_MAXSAMPLEVALUE:
		case TIFFTAG_BITSPERSAMPLE:
			if (!TIFFFetchPerSampleShorts(tif, dp, &v) ||
			    !TIFFSetField(tif, dp->tdir_tag, v))
				goto bad;
			break;
		case TIFFTAG_STRIPOFFSETS:
		case TIFFTAG_STRIPBYTECOUNTS:
			if (td->td_strips == NULL) {
				td->td_strips = (struct strip *)malloc(
				    td->td_nstrips * sizeof (struct strip));
				if (td->td_strips == NULL)
					goto bad;
			}
			if (dp->tdir_tag == TIFFTAG_STRIPOFFSETS) {
				if (!TIFFFetchStripThing(tif, dp,
				    td->td_nstrips,
				    (struct strip *)&td->td_strips->offset))
					goto bad;
				TIFFSetFieldBit(td, FIELD_STRIPOFFSETS);
			} else {
				if (!TIFFFetchStripThing(tif, dp,
				    td->td_nstrips,
				    (struct strip *)&td->td_strips->bytecount))
					goto bad;
				TIFFSetFieldBit(td, FIELD_STRIPBYTECOUNTS);
			}
			break;
		case TIFFTAG_IMAGELENGTH:
		case TIFFTAG_ROWSPERSTRIP:
		case TIFFTAG_SAMPLESPERPIXEL:
		case TIFFTAG_PLANARCONFIG:
			/* handled in first pass above */
			break;
		case TIFFTAG_GRAYRESPONSECURVE:
		case TIFFTAG_COLORRESPONSECURVE:
		case TIFFTAG_COLORMAP:
			v = (1<<td->td_bitspersample) * sizeof (u_short);
			cp = malloc(dp->tdir_tag == TIFFTAG_GRAYRESPONSECURVE ? 
				v : 3*v);
			if (cp == NULL)
				goto bad;
			if (!TIFFSetField(tif, dp->tdir_tag, cp,cp+v,cp+2*v) ||
			    !TIFFFetchData(tif, dp, cp))
				goto bad;
			break;
/* BEGIN REV 4.0 COMPATIBILITY */
		case TIFFTAG_OSUBFILETYPE:
			v = 0;
			switch (TIFFExtractData(tif, dp->tdir_type,
			    dp->tdir_offset)) {
			case OFILETYPE_REDUCEDIMAGE:
				v = FILETYPE_REDUCEDIMAGE;
				break;
			case OFILETYPE_PAGE:
				v = FILETYPE_PAGE;
				break;
			}
			if (!TIFFSetField(tif, dp->tdir_tag, v))
				goto bad;
			break;
/* END REV 4.0 COMPATIBILITY */
		default:
			if (!TIFFSetField(tif, dp->tdir_tag,
			  TIFFExtractData(tif, dp->tdir_type, dp->tdir_offset)))
				goto bad;
			break;
		}
#ifdef BYTECOUNTHACK
	if (!TIFFFieldSet(td, FIELD_STRIPBYTECOUNTS)) {
		/*
		 * Some manufacturers violate the spec by not giving
		 * the size of the strips.  In this case, assume there
		 * is one uncompressed strip of data.
		 */
		if ((TIFFFieldSet(td, FIELD_COMPRESSION) &&
		     td->td_compression != COMPRESSION_NONE) ||
		    (TIFFFieldSet(td, FIELD_ROWSPERSTRIP) &&
		     td->td_rowsperstrip >= td->td_imagelength)) {
			TIFFError(tif->tif_name,
	    "TIFF directory is missing required \"%s\" field",
			    FieldWithTag(TIFFTAG_STRIPBYTECOUNTS)->field_name);
			goto bad;
		}
		TIFFWarning(tif->tif_name,
"TIFF directory is missing required \"%s\" field, calculating from imagelength",
		    FieldWithTag(TIFFTAG_STRIPBYTECOUNTS)->field_name);
		td->td_strips[0].bytecount = td->td_bitspersample *
		    td->td_imagelength * td->td_imagewidth/8;
		TIFFSetFieldBit(td, FIELD_STRIPBYTECOUNTS);
		if (!TIFFFieldSet(td, FIELD_ROWSPERSTRIP))
			td->td_rowsperstrip = td->td_imagelength;
	}
#endif
	if (dir)
		free((char *)dir);
	if (!TIFFFieldSet(td, FIELD_MAXSAMPLEVALUE))
		td->td_maxsamplevalue = (1<<td->td_bitspersample)-1;
	/*
	 * Setup default compression scheme.
	 */
	return (!TIFFFieldSet(td, FIELD_COMPRESSION) ?
	    TIFFSetField(tif, TIFFTAG_COMPRESSION, COMPRESSION_NONE) : 1);
bad:
	if (dir)
		free((char *)dir);
	return (0);
}

/*
 * Fetch a contiguous directory item.
 */
static
TIFFFetchData(tif, dir, cp)
	TIFF *tif;
	TIFFDirEntry *dir;
	char *cp;
{
	int cc, w;

	w = datawidth[dir->tdir_type];
	cc = dir->tdir_count * w;
	if (SeekOK(tif->tif_fd, dir->tdir_offset, L_SET) &&
	    ReadOK(tif->tif_fd, cp, cc)) {
		if (tif->tif_flags & TIFF_SWAB) {
			switch (dir->tdir_type) {
			case SHORT:
				TIFFSwabArrayOfShort(cp, dir->tdir_count);
				break;
			case LONG:
				TIFFSwabArrayOfLong(cp, dir->tdir_count);
				break;
			case RATIONAL:
				TIFFSwabArrayOfLong(cp, 2*dir->tdir_count);
				break;
#ifdef notdef
			case IEEEFLOAT:
				TIFFSwabArrayOfFloat(cp, dir->tdir_count);
				break;
			case IEEEDOUBLE:
				TIFFSwabArrayOfDouble(cp, dir->tdir_count);
				break;
#endif
			}
		}
		return (cc);
	}
	TIFFError(tif->tif_name, "Error fetching data for field \"%s\"",
	    FieldWithTag(dir->tdir_tag)->field_name);
	return (0);
}

/*
 * Fetch a rational item from the file
 * at offset off.  We return the value
 * as floating point number.
 */
static double
TIFFFetchRational(tif, dir)
	TIFF *tif;
	TIFFDirEntry *dir;
{
	long l[2];

	if (!TIFFFetchData(tif, dir, l))
		return (1.);
	if (l[1] == 0) {
		TIFFError(tif->tif_name, "%s: Rational with zero denominator",
		    FieldWithTag(dir->tdir_tag)->field_name);
		return (1.);
	}
	return ((double)l[0] / (double)l[1]);
}

#ifdef notdef
/*
 * Fetch an IEEE floating point item from the file
 * at offset off.  We return the value
 * as floating point number.
 */
static double
TIFFFetchIEEE(tif, dir)
	TIFF *tif;
	TIFFDirEntry *dir;
{
	float f;
	double d;

	if (dir->tdir_type == IEEEFLOAT)
		return (TIFFFetchData(tif, dir, &f) ? f : 1.);
	else
		return (TIFFFetchData(tif, dir, &d) ? d : 1.);
}
#endif

static
TIFFFetchPerSampleShorts(tif, dir, pl)
	TIFF *tif;
	TIFFDirEntry *dir;
	long *pl;
{
	u_short v[4];
	int i, samplesperpixel = tif->tif_dir->td_samplesperpixel;

	if (samplesperpixel <= 2) {
		if (tif->tif_header.tiff_magic == TIFF_BIGENDIAN) {
			v[0] = dir->tdir_offset >> 16;
			v[1] = dir->tdir_offset & 0xffff;
		} else {
			v[0] = dir->tdir_offset & 0xffff;
			v[1] = dir->tdir_offset >> 16;
		}
	} else {
		if (!TIFFFetchData(tif, dir, v))
			return (0);
	}
	for (i = 1; i < samplesperpixel; i++)
		if (v[i] != v[0]) {
			TIFFError(tif->tif_name,
	"Cannot handle different per-sample values for field \"%s\"",
			     FieldWithTag(dir->tdir_tag)->field_name);
			return (0);
		}
ok:
	*pl = v[0];
	return (1);
}

static
TIFFFetchStripThing(tif, dir, nstrips, sp)
	TIFF *tif;
	TIFFDirEntry *dir;
	int nstrips;
	register struct strip *sp;
{
	char *dp;
	int status, space;

	if (nstrips != dir->tdir_count) {
		TIFFError(tif->tif_name,
		    "Count mismatch for field \"%s\"; expecting %d, got %d",
		    FieldWithTag(dir->tdir_tag)->field_name, nstrips,
		    dir->tdir_count);
		return (0);
	}
	space = dir->tdir_count * datawidth[dir->tdir_type];
	if (space <= sizeof (u_long)) {
		/*
		 * Extract data from offset field.
		 */
		 if (dir->tdir_type == (int)SHORT) {
			if (tif->tif_header.tiff_magic == TIFF_BIGENDIAN) {
				(sp++)->offset = dir->tdir_offset >> 16;
				sp->offset = dir->tdir_offset & 0xffff;
			} else {
				(sp++)->offset = dir->tdir_offset & 0xffff;
				sp->offset = dir->tdir_offset >> 16;
			}
		 } else
			sp->offset = dir->tdir_offset;
		status = 1;
	} else {
		dp = malloc(space);
		if (dp == NULL) {
			TIFFError(tif->tif_name,
			    "No memory to fetch field \"%s\"",
			    FieldWithTag(dir->tdir_tag)->field_name);
			return (0);
		}
		if (status = TIFFFetchData(tif, dir, dp)) {
			if (dir->tdir_type == (int)SHORT) {
				register u_short *wp = (u_short *)dp;
				while (nstrips-- > 0)
					(sp++)->offset = *wp++;
			} else {
				register u_long *lp = (u_long *)dp;
				while (nstrips-- > 0)
					(sp++)->offset = *lp++;
			}
		}
		free(dp);
	}
	return (status);
}

static TIFFFieldInfo *
FieldWithTag(tag)
	u_short tag;
{
	register TIFFFieldInfo *fip;

	for (fip = FieldInfo; fip < &FieldInfo[NFIELDINFO]; fip++)
		if (fip->field_tag == tag)
			return (fip);
	TIFFError("FieldWithTag", "Internal error, unknown tag 0x%x", tag);
	exit(-1);
}

TIFFFreeDirectory(td)
	register TIFFDirectory *td;
{

	if (TIFFFieldSet(td, FIELD_GRAYRESPONSECURVE))
		free((char *)td->td_grayresponsecurve);
	if (TIFFFieldSet(td, FIELD_COLORRESPONSECURVE)) {
		free((char *)td->td_redresponsecurve);
		free((char *)td->td_blueresponsecurve);
		free((char *)td->td_greenresponsecurve);
	}
	if (TIFFFieldSet(td, FIELD_COLORMAP)) {
		free((char *)td->td_redcolormap);
		free((char *)td->td_bluecolormap);
		free((char *)td->td_greencolormap);
	}
	if (TIFFFieldSet(td, FIELD_DOCUMENTNAME))
		free(td->td_documentname);
	if (TIFFFieldSet(td, FIELD_ARTIST))
		free(td->td_artist);
	if (TIFFFieldSet(td, FIELD_DATETIME))
		free(td->td_datetime);
	if (TIFFFieldSet(td, FIELD_HOSTCOMPUTER))
		free(td->td_hostcomputer);
	if (TIFFFieldSet(td, FIELD_IMAGEDESCRIPTION))
		free(td->td_imagedescription);
	if (TIFFFieldSet(td, FIELD_MAKE))
		free(td->td_make);
	if (TIFFFieldSet(td, FIELD_MODEL))
		free(td->td_model);
	if (TIFFFieldSet(td, FIELD_SOFTWARE))
		free(td->td_software);
	if (TIFFFieldSet(td, FIELD_PAGENAME))
		free(td->td_pagename);
	if  (td->td_strips)
		free((char *)td->td_strips);
	free((char *)td);
}

static TIFFDirectory *
TIFFNewDirectory(tif)
	register TIFF *tif;
{

	if (tif->tif_dir)
		TIFFFreeDirectory(tif->tif_dir);
	tif->tif_dir = (TIFFDirectory *)malloc(sizeof (TIFFDirectory));
	if (tif->tif_dir == NULL)
		TIFFError(tif->tif_name,
		    "No space for TIFF directory structure");
	else
		bzero((char *)tif->tif_dir, sizeof (TIFFDirectory));
	return (tif->tif_dir);
}

/*
 * Setup a default directory structure.
 */
TIFFDefaultDirectory(tif)
	TIFF *tif;
{
	register TIFFDirectory *td;

	td = TIFFNewDirectory(tif);
	if (td == NULL)
		return (0);
	td->td_bitspersample = 1;
	td->td_threshholding = THRESHHOLD_BILEVEL;
	td->td_orientation = ORIENTATION_TOPLEFT;
	td->td_samplesperpixel = 1;
	td->td_rowsperstrip = 0xfffffff;
	td->td_grayresponseunit = GRAYRESPONSEUNIT_100S;
	td->td_resolutionunit = RESUNIT_INCH;
	td->td_colorresponseunit = COLORRESPONSEUNIT_100S;
	return (TIFFSetField(tif, TIFFTAG_COMPRESSION, COMPRESSION_NONE));
}

/*
 * Record the value of a field in the
 * internal directory structure.  The
 * field will be written to the file
 * when/if the directory structure is
 * updated.
 */
/*VARARGS2*/
TIFFSetField(tif, tag, va_alist)
	TIFF *tif;
	u_short tag;
	va_dcl
{
	static char module[] = "TIFFSetField";
	TIFFDirectory *td = tif->tif_dir;
	va_list ap;
	long v;
	int field = -1, status = 1;

	if (tag != TIFFTAG_IMAGELENGTH && (tif->tif_flags & TIFF_BEENWRITING)) {
		TIFFError(module,
		    "%s: Cannot change TIFF directory while writing",
		    tif->tif_name);
		return (0);
	}
	va_start(ap);
	if (tif->tif_setfield)
		status = (*tif->tif_setfield)(tif, tag, ap);
	if (status) switch (tag) {
	case TIFFTAG_SUBFILETYPE:
		td->td_subfiletype = va_arg(ap, int);
		field = FIELD_SUBFILETYPE;
		break;
	case TIFFTAG_IMAGEWIDTH:
		td->td_imagewidth = va_arg(ap, int);
		field = FIELD_IMAGEDIMENSIONS;
		break;
	case TIFFTAG_IMAGELENGTH:
		td->td_imagelength = va_arg(ap, int);
		field = FIELD_IMAGEDIMENSIONS;
		break;
	case TIFFTAG_BITSPERSAMPLE:
		td->td_bitspersample = va_arg(ap, int);
		field = FIELD_BITSPERSAMPLE;
		break;
	case TIFFTAG_COMPRESSION:
		v = va_arg(ap, int) & 0xffff;
		/*
		 * If we're changing the compression scheme,
		 * the notify the previous module so that it
		 * can cleanup any state it's setup.
		 */
		if (TIFFFieldSet(td, FIELD_COMPRESSION)) {
			if (td->td_compression == v)
				break;
			if (tif->tif_done)
				(*tif->tif_done)(tif);
		}
		/*
		 * Setup new compression routine state.
		 */
		if (status = TIFFSetCompressionScheme(tif, v)) {
			td->td_compression = v;
			field = FIELD_COMPRESSION;
		}
		break;
	case TIFFTAG_PHOTOMETRIC:
		td->td_photometric = va_arg(ap, int);
		field = FIELD_PHOTOMETRIC;
		break;
	case TIFFTAG_THRESHHOLDING:
		td->td_threshholding = va_arg(ap, int);
		field = FIELD_THRESHHOLDING;
		break;
	case TIFFTAG_DOCUMENTNAME:
		td->td_documentname = va_arg(ap, char *);
		field = FIELD_DOCUMENTNAME;
		break;
	case TIFFTAG_ARTIST:
		td->td_artist = va_arg(ap, char *);
		field = FIELD_ARTIST;
		break;
	case TIFFTAG_DATETIME:
		td->td_datetime = va_arg(ap, char *);
		field = FIELD_DATETIME;
		break;
	case TIFFTAG_HOSTCOMPUTER:
		td->td_hostcomputer = va_arg(ap, char *);
		field = FIELD_HOSTCOMPUTER;
		break;
	case TIFFTAG_IMAGEDESCRIPTION:
		td->td_imagedescription = va_arg(ap, char *);
		field = FIELD_IMAGEDESCRIPTION;
		break;
	case TIFFTAG_MAKE:
		td->td_make = va_arg(ap, char *);
		field = FIELD_MAKE;
		break;
	case TIFFTAG_MODEL:
		td->td_model = va_arg(ap, char *);
		field = FIELD_MODEL;
		break;
	case TIFFTAG_SOFTWARE:
		td->td_software = va_arg(ap, char *);
		field = FIELD_SOFTWARE;
		break;
	case TIFFTAG_ORIENTATION:
		td->td_orientation = va_arg(ap, int);
		field = FIELD_ORIENTATION;
		break;
	case TIFFTAG_SAMPLESPERPIXEL:
		v = va_arg(ap, int);
		if (v == 0)
			goto badvalue;
		if (v > 4) {
			TIFFError(tif->tif_name,
			    "Cannot handle %d-channel data", v); 
			goto bad;
		}
		td->td_samplesperpixel = v;
		field = FIELD_SAMPLESPERPIXEL;
		break;
	case TIFFTAG_ROWSPERSTRIP:
		v = va_arg(ap, long);
		if (v == 0)
			goto badvalue;
		td->td_rowsperstrip = v;
		field = FIELD_ROWSPERSTRIP;
		break;
	case TIFFTAG_MINSAMPLEVALUE:
		td->td_minsamplevalue = va_arg(ap, int) & 0xffff;
		field = FIELD_MINSAMPLEVALUE;
		break;
	case TIFFTAG_MAXSAMPLEVALUE:
		td->td_maxsamplevalue = va_arg(ap, int) & 0xffff;
		field = FIELD_MAXSAMPLEVALUE;
		break;
	case TIFFTAG_XRESOLUTION:
		td->td_xresolution = va_arg(ap, double);
		field = FIELD_RESOLUTION;
		break;
	case TIFFTAG_YRESOLUTION:
		td->td_yresolution = va_arg(ap, double);
		field = FIELD_RESOLUTION;
		break;
	case TIFFTAG_PLANARCONFIG:
		v = va_arg(ap, int);
		if (v != PLANARCONFIG_CONTIG) {		/* XXX */
			TIFFError(tif->tif_name,
		     "Sorry, can only handle contiguous data (\"%s\"=%d)",
			    FieldWithTag(tag)->field_name, v);
			goto bad;
		}
		td->td_planarconfig = v;
		field = FIELD_PLANARCONFIG;
		break;
	case TIFFTAG_PAGENAME:
		td->td_pagename = va_arg(ap, char *);
		field = FIELD_PAGENAME;
		break;
	case TIFFTAG_XPOSITION:
		td->td_xposition = va_arg(ap, double);
		field = FIELD_POSITION;
		break;
	case TIFFTAG_YPOSITION:
		td->td_yposition = va_arg(ap, double);
		field = FIELD_POSITION;
		break;
	case TIFFTAG_GRAYRESPONSEUNIT:
		td->td_grayresponseunit = va_arg(ap, int);
		field = FIELD_GRAYRESPONSEUNIT;
		break;
	case TIFFTAG_GRAYRESPONSECURVE:
		td->td_grayresponsecurve = va_arg(ap, u_short *);
		field = FIELD_GRAYRESPONSECURVE;
		break;
	case TIFFTAG_GROUP3OPTIONS:
		td->td_group3options = va_arg(ap, long);
		field = FIELD_GROUP3OPTIONS;
		break;
	case TIFFTAG_GROUP4OPTIONS:
		td->td_group4options = va_arg(ap, long);
		field = FIELD_GROUP4OPTIONS;
		break;
	case TIFFTAG_RESOLUTIONUNIT:
		td->td_resolutionunit = va_arg(ap, int);
		field = FIELD_RESOLUTIONUNIT;
		break;
	case TIFFTAG_PAGENUMBER:
		td->td_pagenumber = va_arg(ap, int);
		field = FIELD_PAGENUMBER;
		break;
	case TIFFTAG_COLORRESPONSEUNIT:
		td->td_colorresponseunit = va_arg(ap, int);
		field = FIELD_COLORRESPONSEUNIT;
		break;
	case TIFFTAG_COLORRESPONSECURVE:
		td->td_redresponsecurve = va_arg(ap, u_short *);
		td->td_greenresponsecurve = va_arg(ap, u_short *);
		td->td_blueresponsecurve = va_arg(ap, u_short *);
		field = FIELD_COLORRESPONSECURVE;
		break;
	case TIFFTAG_COLORMAP:
		td->td_redcolormap = va_arg(ap, u_short *);
		td->td_greencolormap = va_arg(ap, u_short *);
		td->td_bluecolormap = va_arg(ap, u_short *);
		field = FIELD_COLORMAP;
		break;
	case TIFFTAG_PREDICTOR:
		td->td_predictor = va_arg(ap, int);
		field = FIELD_PREDICTOR;
		break;
	case TIFFTAG_MATTEING:
		td->td_matteing = va_arg(ap, int);
		field = FIELD_MATTEING;
		break;
	}
	if (field >= 0) {
		TIFFSetFieldBit(td, field);
		tif->tif_flags |= TIFF_DIRTYDIRECT;
	}
	va_end(ap);
	return (status);
badvalue:
	TIFFError(tif->tif_name, "%d: Bad value for \"%s\"", v,
	    FieldWithTag(tag)->field_name);
bad:
	va_end(ap);
	return (0);
}

/*
 * Return the value of a field in the
 * internal directory structure.
 */
TIFFGetField(tif, tag, va_alist)
	TIFF *tif;
	u_short tag;
	va_dcl
{
	register TIFFFieldInfo *fip;
	TIFFDirectory *td = tif->tif_dir;
	va_list ap;
	int status;

	for (fip = FieldInfo; fip < &FieldInfo[NFIELDINFO]; fip++)
		if (fip->field_tag == tag)
			break;
	if (fip >= &FieldInfo[NFIELDINFO]) {
		TIFFError("TIFFGetField", "Unknown field, tag 0x%x", tag);
		return (0);
	}
	if (TIFFFieldSet(td, fip->field_bit)) {
		va_start(ap);
		(void) TIFFGetField1(td, tag, ap);
		va_end(ap);
		return (1);
	}
	return (0);
}

static
TIFFGetField1(td, tag, ap)
	TIFFDirectory *td;
	u_short tag;
	va_list ap;
{

	switch (tag) {
	case TIFFTAG_SUBFILETYPE:
		*va_arg(ap, u_short *) = td->td_subfiletype;
		break;
	case TIFFTAG_IMAGEWIDTH:
		*va_arg(ap, u_short *) = td->td_imagewidth;
		break;
	case TIFFTAG_IMAGELENGTH:
		*va_arg(ap, u_short *) = td->td_imagelength;
		break;
	case TIFFTAG_BITSPERSAMPLE:
		*va_arg(ap, u_short *) = td->td_bitspersample;
		break;
	case TIFFTAG_COMPRESSION:
		*va_arg(ap, u_short *) = td->td_compression;
		break;
	case TIFFTAG_PHOTOMETRIC:
		*va_arg(ap, u_short *) = td->td_photometric;
		break;
	case TIFFTAG_THRESHHOLDING:
		*va_arg(ap, u_short *) = td->td_threshholding;
		break;
	case TIFFTAG_DOCUMENTNAME:
		*va_arg(ap, char **) = td->td_documentname;
		break;
	case TIFFTAG_ARTIST:
		*va_arg(ap, char **) = td->td_artist;
		break;
	case TIFFTAG_DATETIME:
		*va_arg(ap, char **) = td->td_datetime;
		break;
	case TIFFTAG_HOSTCOMPUTER:
		*va_arg(ap, char **) = td->td_hostcomputer;
		break;
	case TIFFTAG_IMAGEDESCRIPTION:
		*va_arg(ap, char **) = td->td_imagedescription;
		break;
	case TIFFTAG_MAKE:
		*va_arg(ap, char **) = td->td_make;
		break;
	case TIFFTAG_MODEL:
		*va_arg(ap, char **) = td->td_model;
		break;
	case TIFFTAG_SOFTWARE:
		*va_arg(ap, char **) = td->td_software;
		break;
	case TIFFTAG_ORIENTATION:
		*va_arg(ap, u_short *) = td->td_orientation;
		break;
	case TIFFTAG_SAMPLESPERPIXEL:
		*va_arg(ap, u_short *) = td->td_samplesperpixel;
		break;
	case TIFFTAG_ROWSPERSTRIP:
		*va_arg(ap, u_long *) = td->td_rowsperstrip;
		break;
	case TIFFTAG_MINSAMPLEVALUE:
		*va_arg(ap, u_short *) = td->td_minsamplevalue;
		break;
	case TIFFTAG_MAXSAMPLEVALUE:
		*va_arg(ap, u_short *) = td->td_maxsamplevalue;
		break;
	case TIFFTAG_XRESOLUTION:
		*va_arg(ap, float *) = td->td_xresolution;
		break;
	case TIFFTAG_YRESOLUTION:
		*va_arg(ap, float *) = td->td_yresolution;
		break;
	case TIFFTAG_PLANARCONFIG:
		*va_arg(ap, u_short *) = td->td_planarconfig;
		break;
	case TIFFTAG_XPOSITION:
		*va_arg(ap, float *) = td->td_xposition;
		break;
	case TIFFTAG_YPOSITION:
		*va_arg(ap, float *) = td->td_yposition;
		break;
	case TIFFTAG_PAGENAME:
		*va_arg(ap, char **) = td->td_pagename;
		break;
	case TIFFTAG_GRAYRESPONSEUNIT:
		*va_arg(ap, u_short *) = td->td_grayresponseunit;
		break;
	case TIFFTAG_GRAYRESPONSECURVE:
		*va_arg(ap, u_short **) = td->td_grayresponsecurve;
		break;
	case TIFFTAG_GROUP3OPTIONS:
		*va_arg(ap, u_long *) = td->td_group3options;
		break;
	case TIFFTAG_GROUP4OPTIONS:
		*va_arg(ap, u_long *) = td->td_group4options;
		break;
	case TIFFTAG_RESOLUTIONUNIT:
		*va_arg(ap, u_short *) = td->td_resolutionunit;
		break;
	case TIFFTAG_PAGENUMBER:
		*va_arg(ap, u_long *) = td->td_pagenumber;
		break;
	case TIFFTAG_COLORRESPONSEUNIT:
		*va_arg(ap, u_short *) = td->td_colorresponseunit;
		break;
	case TIFFTAG_COLORRESPONSECURVE:
		*va_arg(ap, u_short **) = td->td_redresponsecurve;
		*va_arg(ap, u_short **) = td->td_greenresponsecurve;
		*va_arg(ap, u_short **) = td->td_blueresponsecurve;
		break;
	case TIFFTAG_COLORMAP:
		*va_arg(ap, u_short **) = td->td_redcolormap;
		*va_arg(ap, u_short **) = td->td_greencolormap;
		*va_arg(ap, u_short **) = td->td_bluecolormap;
		break;
	case TIFFTAG_PREDICTOR:
		*va_arg(ap, u_short *) = td->td_predictor;
		break;
	case TIFFTAG_MATTEING:
		*va_arg(ap, u_short *) = td->td_matteing;
		break;
	}
	va_end(ap);
}

/*
 * Internal interface to TIFFGetField...
 */
static
TIFFgetfield(td, tag, va_alist)
	TIFFDirectory *td;
	u_short tag;
	va_dcl
{
	va_list ap;

	va_start(ap);
	(void) TIFFGetField1(td, tag, ap);
	va_end(ap);
}

/* shorthands for setting up and writing directory... */
#define	MakeShortDirent(tag, v) \
	dir->tdir_tag = tag; \
	dir->tdir_type = (short)SHORT; \
	dir->tdir_count = 1; \
	dir->tdir_offset = TIFFInsertData(tif, (int)SHORT, v); \
	dir++
#define	WriteRationalPair(tag1, v1, tag2, v2) \
	(TIFFWriteRational(tif, tag1, dir++, v1) && \
	 TIFFWriteRational(tif, tag2, dir++, v2))

static	long dataoff;
/*
 * Write the contents of the current directory
 * to the specified file.  This routine doesn't
 * handle overwriting a directory with auxiliary
 * storage that's been changed.
 */
TIFFWriteDirectory(tif)
	TIFF *tif;
{
	short dircount, v;
	int nfields, dirsize;
	char *data, *cp;
	TIFFFieldInfo *fip;
	TIFFDirEntry *dir;
	TIFFDirectory *td;
	u_long b, off, fields[sizeof (td->td_fieldsset) / sizeof (u_long)];

	if (tif->tif_mode == O_RDONLY)
		return (1);
	td = tif->tif_dir;
	/*
	 * Size the directory so that we can calculate
	 * offsets for the data items that aren't kept
	 * in-place in each field.
	 */
	nfields = 0;
	for (b = 0; b <= FIELD_LAST; b++)
		if (TIFFFieldSet(td, b))
			nfields += (b < FIELD_SUBFILETYPE ? 2 : 1);
	dirsize = nfields * sizeof (TIFFDirEntry);
	data = malloc(dirsize);
	if (data == NULL) {
		TIFFError(tif->tif_name,
		    "Cannot write directory, out of space");
		return (0);
	}
	/*
	 * Directory hasn't been placed yet, put
	 * it at the end of the file and link it
	 * into the existing directory structure.
	 */
	if (tif->tif_diroff == 0 && !TIFFLinkDirectory(tif))
		return (0);
	dataoff = tif->tif_diroff + sizeof (short) + dirsize + sizeof (long);
	if (dataoff & 1)
		dataoff++;
	(void) lseek(tif->tif_fd, dataoff, L_SET);
	dir = (TIFFDirEntry *)data;
	/*
	 * Setup external form of directory
	 * entries and write data items.
	 */
	bcopy(td->td_fieldsset, fields, sizeof (fields));
	for (fip = FieldInfo; fip < &FieldInfo[NFIELDINFO]; fip++) {
		if (fip->field_bit == (u_short)-1 ||
		    !FieldSet(fields, fip->field_bit))
			continue;
		if (fip->field_type == ASCII) {
			TIFFgetfield(td, fip->field_tag, &cp);
			dir->tdir_tag = fip->field_tag;
			dir->tdir_type = (short)ASCII;
			dir->tdir_count = strlen(cp) + 1;
			if (!TIFFWriteData(tif, dir++, cp))
				goto bad;
			ResetFieldBit(fields, fip->field_bit);
			continue;
		}
		switch (fip->field_bit) {
		case FIELD_STRIPOFFSETS:
		case FIELD_STRIPBYTECOUNTS:
			if (!TIFFWriteStripThing(tif,
			    fip->field_tag, dir++, LONG))
				goto bad;
			break;
		case FIELD_GRAYRESPONSECURVE:
			dir->tdir_tag = fip->field_tag;
			dir->tdir_type = (short)SHORT;
			dir->tdir_count = 1<<td->td_bitspersample;
			if (!TIFFWriteData(tif, dir++,td->td_grayresponsecurve))
				goto bad;
			break;
		case FIELD_COLORRESPONSECURVE:
		case FIELD_COLORMAP:
			dir->tdir_tag = fip->field_tag;
			dir->tdir_type = (short)SHORT;
			/* XXX -- yech, fool TIFFWriteData */
			dir->tdir_count = 1<<td->td_bitspersample;
			off = dataoff;
#define	WRITE(x)	TIFFWriteData(tif, dir, x)
			if (fip->field_tag == TIFFTAG_COLORMAP) {
				if (!WRITE(td->td_redcolormap) ||
				    !WRITE(td->td_greencolormap) ||
				    !WRITE(td->td_bluecolormap))
					goto bad;
			} else {
				if (!WRITE(td->td_redresponsecurve) ||
				    !WRITE(td->td_greenresponsecurve) ||
				    !WRITE(td->td_blueresponsecurve))
					goto bad;
			}
#undef WRITE
			dir->tdir_count *= 3;
			dir->tdir_offset = off;
			break;
		case FIELD_IMAGEDIMENSIONS:
			MakeShortDirent(TIFFTAG_IMAGEWIDTH, td->td_imagewidth);
			MakeShortDirent(TIFFTAG_IMAGELENGTH,td->td_imagelength);
			break;
		case FIELD_POSITION:
			if (!WriteRationalPair(
			    TIFFTAG_XPOSITION, td->td_xposition,
			    TIFFTAG_YPOSITION, td->td_yposition))
				goto bad;
			break;
		case FIELD_RESOLUTION:
			if (!WriteRationalPair(
			    TIFFTAG_XRESOLUTION, td->td_xresolution,
			    TIFFTAG_YRESOLUTION, td->td_yresolution))
				goto bad;
			break;
		case FIELD_BITSPERSAMPLE:
		case FIELD_MINSAMPLEVALUE:
		case FIELD_MAXSAMPLEVALUE:
			TIFFgetfield(td, fip->field_tag, &v);
			if (!TIFFWritePerSampleShorts(tif, fip->field_tag,
			    dir++, v))
				goto bad;
			break;
		default:
			dir->tdir_tag = fip->field_tag;
			dir->tdir_type = (short)fip->field_type;
			dir->tdir_count = fip->field_count;
			if (fip->field_type == SHORT) {
				TIFFgetfield(td, fip->field_tag, &v);
				dir->tdir_offset =
				    TIFFInsertData(tif, dir->tdir_type, v);
			} else
				TIFFgetfield(td, fip->field_tag,
				    &dir->tdir_offset);
			dir++;
			break;
		}
		ResetFieldBit(fields, fip->field_bit);
	}
	/*
	 * Write directory.
	 */
	(void) lseek(tif->tif_fd, tif->tif_diroff, L_SET);
	dircount = nfields;
	if (!WriteOK(tif->tif_fd, &dircount, sizeof (short))) {
		TIFFError(tif->tif_name, "Error writing directory count");
		goto bad;
	}
	if (!WriteOK(tif->tif_fd, data, dirsize)) {
		TIFFError(tif->tif_name, "Error writing directory contents");
		goto bad;
	}
	if (!WriteOK(tif->tif_fd, &tif->tif_nextdiroff, sizeof (long))) {
		TIFFError(tif->tif_name, "Error writing directory link");
		goto bad;
	}
	free(data);
	return (1);
bad:
	free(data);
	return (0);
}
#undef MakeShortDirent
#undef WriteRationalPair

static
TIFFWriteRational(tif, tag, dir, v)
	TIFF *tif;
	u_short tag;
	TIFFDirEntry *dir;
	float v;
{
	long t[2];

	dir->tdir_tag = tag;
	dir->tdir_type = (short)RATIONAL;
	dir->tdir_count = 1;
	/* need algorithm to convert ... XXX */
	t[0] = v * 10000.;
	t[1] = 10000;
	return (TIFFWriteData(tif, dir, t));
}

static
TIFFWritePerSampleShorts(tif, tag, dir, v)
	TIFF *tif;
	u_short tag;
	register TIFFDirEntry *dir;
	short v;
{
	short w[4];
	int i, samplesperpixel = tif->tif_dir->td_samplesperpixel;

	dir->tdir_tag = tag;
	dir->tdir_type = (short)SHORT;
	dir->tdir_count = samplesperpixel;
	if (samplesperpixel <= 2) {
		if (tif->tif_header.tiff_magic == TIFF_BIGENDIAN) {
			dir->tdir_offset = v << 16;
			if (samplesperpixel == 2)
				dir->tdir_offset |= v & 0xffff;
		} else {
			dir->tdir_offset = v & 0xffff;
			if (samplesperpixel == 2)
				dir->tdir_offset |= v << 16;
		}
		return (1);
	}
	for (i = 0; i < samplesperpixel; i++)
		w[i] = v;
	return (TIFFWriteData(tif, dir, w));
}

static
TIFFWriteStripThing(tif, tag, dir, type)
	TIFF *tif;
	u_short tag;
	TIFFDirEntry *dir;
	TIFFDataType type;
{
	register struct strip *sp;
	register int i;
	TIFFDirectory *td = tif->tif_dir;
	char *dp;

	dir->tdir_tag = tag;
	dir->tdir_type = (short)type;
	dir->tdir_count = (td->td_rowsperstrip == 0xffffffff ?
	   1 : howmany(td->td_imagelength, td->td_rowsperstrip));
	i = dir->tdir_count * datawidth[(int)type];
	sp = (dir->tdir_tag == TIFFTAG_STRIPOFFSETS ?
	    (struct strip *)&td->td_strips->offset :
	    (struct strip *)&td->td_strips->bytecount);
	if (i <= sizeof (u_long)) {
		/*
		 * Place value(s) in-line in offset field.
		 */
		if (type == SHORT) {
			if (tif->tif_header.tiff_magic == TIFF_BIGENDIAN) {
				dir->tdir_offset = (sp++)->offset << 16;
				if (dir->tdir_count == 2)
					dir->tdir_offset |= sp->offset & 0xffff;
			} else {
				dir->tdir_offset = (sp++)->offset & 0xffff;
				if (dir->tdir_count == 2)
					dir->tdir_offset |= sp->offset << 16;
			}
		} else
			dir->tdir_offset = sp->offset;
		return (1);
	}
	dp = malloc(i);
	if (dp == NULL) {
		TIFFError(tif->tif_name,
		    "Out of memory while writing field \"%s\"",
		    FieldWithTag(dir->tdir_tag)->field_name);
		return (0);
	}
	if (type == SHORT) {
		register u_short *wp = (u_short *)dp;

		for (i = dir->tdir_count; i-- > 0;)
			*wp++ = (sp++)->offset;
	} else {
		register u_long *lp = (u_long *)dp;

		for (i = dir->tdir_count; i-- > 0;)
			*lp++ = (sp++)->offset;
	}
	i = TIFFWriteData(tif, dir, dp);
	free(dp);
	return (i);
}

/*
 * Write a contiguous directory item.
 */
static
TIFFWriteData(tif, dir, cp)
	TIFF *tif;
	TIFFDirEntry *dir;
	char *cp;
{
	int cc;

	dir->tdir_offset = dataoff;
	cc = dir->tdir_count * datawidth[dir->tdir_type];
	if (SeekOK(tif->tif_fd, dir->tdir_offset, L_SET) &&
	    WriteOK(tif->tif_fd, cp, cc)) {
		dataoff += (cc + 1) & ~1;
		return (1);
	}
	TIFFError(tif->tif_name, "Error writing data for field \"%s\"",
	    FieldWithTag(dir->tdir_tag)->field_name);
	return (0);
}

/*
 * Link the current directory into the
 * directory chain for the file.
 */
static
TIFFLinkDirectory(tif)
	register TIFF *tif;
{
	static char module[] = "TIFFLinkDirectory";
	short dircount;
	long nextdir;

	tif->tif_diroff = (lseek(tif->tif_fd, 0L, L_XTND)+1) &~ 1;
	if (tif->tif_header.tiff_diroff == 0) {
		/*
		 * First directory, overwrite header.
		 */
		tif->tif_header.tiff_diroff = tif->tif_diroff;
		(void) lseek(tif->tif_fd, 0L, L_SET);
		if (!WriteOK(tif->tif_fd, &tif->tif_header,
		    sizeof (tif->tif_header))) {
			TIFFError(tif->tif_name, "Error writing TIFF header");
			return (0);
		}
		return (1);
	}
	/*
	 * Not the first directory, search to the last and append.
	 */
	nextdir = tif->tif_header.tiff_diroff;
	do {
		if (!SeekOK(tif->tif_fd, nextdir, L_SET) ||
		    !ReadOK(tif->tif_fd, &dircount, sizeof (dircount))) {
			TIFFError(module, "Error fetching directory count");
			return (0);
		}
		if (!SeekOK(tif->tif_fd, dircount * sizeof (TIFFDirEntry),
			L_INCR) ||
		    !ReadOK(tif->tif_fd, &nextdir, sizeof (nextdir))) {
			TIFFError(module, "Error fetching directory link");
			return (0);
		}
	} while (nextdir != 0);
	(void) lseek(tif->tif_fd, -sizeof (nextdir), L_INCR);
	if (!WriteOK(tif->tif_fd, &tif->tif_diroff, sizeof (tif->tif_diroff))) {
		TIFFError(module, "Error writing directory link");
		return (0);
	}
	return (1);
}

/*
 * Set the n-th directory as the current directory.
 * NB: Directories are numbered starting at 1.
 */
TIFFSetDirectory(tif, n)
	register TIFF *tif;
	int n;
{
	static char module[] = "TIFFSetDirectory";
	short dircount;
	long nextdir;

	nextdir = tif->tif_header.tiff_diroff;
	if (nextdir == 0)
		return (0);
	while (n-- > 0 && nextdir != 0) {
		if (!SeekOK(tif->tif_fd, nextdir, L_SET) ||
		    !ReadOK(tif->tif_fd, &dircount, sizeof (dircount))) {
			TIFFError(module, "%s: Error fetching directory count",
			    tif->tif_name);
			return (0);
		}
		if (!SeekOK(tif->tif_fd, dircount * sizeof (TIFFDirEntry),
			L_INCR) ||
		    !ReadOK(tif->tif_fd, &nextdir, sizeof (nextdir))) {
			TIFFError(module, "%s: Error fetching directory link",
			    tif->tif_name);
			return (0);
		}
	}
	tif->tif_diroff = lseek(tif->tif_fd, -sizeof (nextdir), L_INCR);
	return (TIFFReadDirectory(tif));
}
