#ifndef lint
static char sccsid[] = "@(#)tiffcmp.c	1.3 2/19/89";
#endif

/*
 * Copyright (c) 1988 by Sam Leffler.
 * All rights reserved.
 *
 * This file is provided for unrestricted use provided that this
 * legend is included on all tape media and as a part of the
 * software program in whole or part.  Users may copy, modify or
 * distribute this file at will.
 */

#include <stdio.h>
#include "tiffio.h"

#define	howmany(x, y)	(((x)+((y)-1))/(y))

main(argc, argv)
	char *argv[];
{
	TIFF *tif1, *tif2;
	short bitspersample, imagewidth;
	int size, row;
	register u_char *p1, *p2;
	register int byte;
	u_char *buf1, *buf2;
	int stoponfirstdiff = 1;

	argc--, argv++;
	if (argc > 0 && strcmp(argv[0], "-l") == 0) {
		stoponfirstdiff = 0;
		argc--, argv++;
	}
	if (argc < 2) {
		fprintf(stderr, "Usage: tiffcmp [-l] file1 file2\n");
		exit(-3);
	}
	tif1 = TIFFOpen(argv[0], "r");
	if (tif1 == NULL)
		exit(-1);
	tif2 = TIFFOpen(argv[1], "r");
	if (tif2 == NULL)
		exit(-2);
	if (!CheckTag(tif1, tif2, TIFFTAG_BITSPERSAMPLE, "BitsPerSample"))
		exit(1);
	if (!CheckTag(tif1, tif2, TIFFTAG_SAMPLESPERPIXEL, "SamplesPerPixel"))
		exit(1);
	if (!CheckTag(tif1, tif2, TIFFTAG_IMAGEWIDTH, "ImageWidth"))
		exit(1);
	(void) TIFFGetField(tif1, TIFFTAG_BITSPERSAMPLE, &bitspersample);
	(void) TIFFGetField(tif1, TIFFTAG_IMAGEWIDTH, &imagewidth);
	size = howmany(bitspersample * imagewidth, 8);
	buf1 = (u_char *)malloc(2 * size);
	if (buf1 == NULL) {
		fprintf(stderr, "No space for scanline buffers\n");
		exit(-1);
	}
	buf2 = buf1 + size;
	row = 1;
	while (TIFFReadScanline(tif1, buf1) > 0) {
		if (TIFFReadScanline(tif2, buf2) <= 0) {
			printf("%s: EOF at scanline %d\n", tif2->tif_name, row);
			exit(1);
		}
		byte++;
		for (p1 = buf1, p2 = buf2; byte < size ; p1++, p2++) {
			if (*p1 != *p2) {
				printf("Scanline %d, byte %d: %02x %02x\n",
				    row, *p1, *p2);
				if (stoponfirstdiff)
					exit(1);
			}
			byte++;
		}
		row++;
	}
	if (TIFFReadScanline(tif2, buf2) > 0) {
		printf("%s: EOF at scanline %d\n", tif1->tif_name, row);
		exit(1);
	}
	exit(0);
}

CheckTag(tif1, tif2, tag, name)
	TIFF *tif1, *tif2;
	int tag;
	char *name;
{
	short v1, v2;

	if (TIFFGetField(tif1, tag, &v1)) {
		if (!TIFFGetField(tif2, tag, &v2)) {
			printf("Tag \"%s\" appears only in %s\n",
			    name, tif1->tif_name);
			return (0);
		}
	} else if (TIFFGetField(tif2, tag, &v2)) {
		printf("Tag \"%s\" appears only in %s\n",
		    name, tif2->tif_name);
		return (0);
	}
	if (v1 != v2) {
		printf("Difference for tag \"%s\" (%d != %d)\n", name, v1, v2);
		return (0);
	}
	return (1);
}
