#ifndef lint
static char sccsid[] = "@(#)tiffcp.c	1.2 2/19/89";
#endif

/*
 * Copyright (c) 1988 by Sam Leffler.
 * All rights reserved.
 *
 * This file is provided for unrestricted use provided that this
 * legend is included on all tape media and as a part of the
 * software program in whole or part.  Users may copy, modify or
 * distribute this file at will.
 */

#include <stdio.h>
#include "tiffio.h"

#define	howmany(x, y)	(((x)+((y)-1))/(y))
#define	streq(a,b)	(strcmp(a,b) == 0)
#define	CopyField(tag, v) \
	if (TIFFGetField(in, tag, &v)) TIFFSetField(out, tag, v)

main(argc, argv)
	char *argv[];
{
	register TIFF *in, *out;
	short bitspersample, samplesperpixel, shortv;
	short imagewidth;
	u_short compression = -1;
	long rowsperstrip = -1;
	float floatv;
	u_char *buf;

	argc--, argv++;
	if (argc < 2)
		usage();
	for (; argc > 2 && argv[0][0] == '-'; argc--, argv++) {
		if (streq(argv[0], "-picio")) {
			compression = COMPRESSION_PICIO;
			continue;
		}
		if (streq(argv[0], "-none")) {
			compression = COMPRESSION_NONE;
			continue;
		}
		if (streq(argv[0], "-packbits")) {
			compression = COMPRESSION_PACKBITS;
			continue;
		}
		if (streq(argv[0], "-thunderscan")) {
			compression = COMPRESSION_THUNDERSCAN;
			continue;
		}
		if (streq(argv[0], "-lzw")) {
			compression = COMPRESSION_LZW;
			continue;
		}
		if (streq(argv[0], "-rowsperstrip")) {
			argc--, argv++;
			rowsperstrip = atoi(argv[0]);
			continue;
		}
		usage();
	}
	in = TIFFOpen(argv[0], "r");
	if (in == NULL)
		exit(-1);
	out = TIFFOpen(argv[1], "w");
	if (out == NULL)
		exit(-2);
	CopyField(TIFFTAG_SUBFILETYPE, shortv);
	CopyField(TIFFTAG_IMAGEWIDTH, imagewidth);
	CopyField(TIFFTAG_IMAGELENGTH, shortv);
	CopyField(TIFFTAG_BITSPERSAMPLE, bitspersample);
	if (compression != (u_short)-1)
		TIFFSetField(out, TIFFTAG_COMPRESSION, compression);
	else
		CopyField(TIFFTAG_COMPRESSION, compression);
	CopyField(TIFFTAG_PHOTOMETRIC, shortv);
	CopyField(TIFFTAG_THRESHHOLDING, shortv);
	CopyField(TIFFTAG_ORIENTATION, shortv);
	CopyField(TIFFTAG_SAMPLESPERPIXEL, samplesperpixel);
	if (rowsperstrip <= 0) {
		CopyField(TIFFTAG_ROWSPERSTRIP, rowsperstrip);
	} else
		TIFFSetField(out, TIFFTAG_ROWSPERSTRIP, rowsperstrip);
	CopyField(TIFFTAG_MINSAMPLEVALUE, shortv);
	CopyField(TIFFTAG_MAXSAMPLEVALUE, shortv);
	CopyField(TIFFTAG_XRESOLUTION, floatv);
	CopyField(TIFFTAG_YRESOLUTION, floatv);
	CopyField(TIFFTAG_PLANARCONFIG, shortv);
	if (shortv == PLANARCONFIG_CONTIG)
		bitspersample *= samplesperpixel;
	buf = (u_char *)malloc(howmany(bitspersample * imagewidth, 8));
	if (buf == NULL) {
		fprintf(stderr, "No space for scanline buffer\n");
		exit(-1);
	}
	CopyField(TIFFTAG_XPOSITION, floatv);
	CopyField(TIFFTAG_YPOSITION, floatv);
	/* XXX what about response curves? */
	CopyField(TIFFTAG_MATTEING, shortv);
	while (TIFFReadScanline(in, buf) > 0)
		if (TIFFWriteScanline(out, buf) < 0)
			break;
	(void) TIFFClose(in);
	(void) TIFFClose(out);
}

usage()
{
	fprintf(stderr, "usage: tiffcp [options] input output\n");
	fprintf(stderr, "where options are:\n");
	fprintf(stderr,
	    " -picio\t\tcompress output with picio encoding\n");
	fprintf(stderr,
	    " -lzw\t\tcompress output with Lempel-Ziv & Welch encoding\n");
	fprintf(stderr,
	    " -packbits\tcompress output with packbits encoding\n");
	fprintf(stderr,
	    " -thunderscan\tcompress output with ThunderScan encoding\n");
	fprintf(stderr,
	    " -none\t\tuse no compression algorithm on output\n");
	fprintf(stderr, "\n");
	fprintf(stderr,
	    " -rowsperstrip #\tmake each strip have no more than # rows\n");
	exit(-1);
}
